/*
 * spec : root : root of Engrave
 * gl: context (not required)
 */
var Engrave=(function(){    
    var GL, JQCV, CV, VBO, VBOI, SPEC, TEXTURE,
        FB,RB,TEXTURE_BW,
        FB_AO, TEXTURE_AO,
        FB_NORMAL, TEXTURE_NORMAL, RB2,
        X,Y,W,H,X_SIZE, Y_SIZE,
        SHADER_PROGRAMS={bw: false,
                 ao: false,
                 displacement: false,
                 normal: false};

    var get_shaders=function(vertex_source, fragment_source, id, callback, clearColor){
        var shaders={
            vertex: false,
            fragment: false
        }
        var get_shader=function(url, type, str) {
            $.ajax({
             url: SPEC.root+url,
             dataType: "text",
             context: document.body,
             async: true
            }).success(function(source){
                var shader = GL.createShader(type);
                GL.shaderSource(shader, source);
                GL.compileShader(shader);
                if (!GL.getShaderParameter(shader, GL.COMPILE_STATUS)) {
                    alert("Error in "+id+" "+str+" shader : \n"+GL.getShaderInfoLog(shader));
                    return false;
                }
                shaders[str]=shader;
                if (shaders.vertex && shaders.fragment) linkShaders();
            });
            
        };
        
        var linkShaders=function() {           
            var shader_program=GL.createProgram();
            GL.attachShader(shader_program, shaders.vertex);
            GL.attachShader(shader_program, shaders.fragment);
            GL.linkProgram(shader_program);
            SHADER_PROGRAMS[id]=shader_program;
            shader_program.clearColor=clearColor;
            callback(shader_program);
        }
        get_shader(vertex_source, GL.VERTEX_SHADER, "vertex"),
        get_shader(fragment_source, GL.FRAGMENT_SHADER, "fragment");
    }

    var makeImage=function(shader_program, texture, setUniforms, fullViewPort, fb){
        GL.useProgram(shader_program);
        GL.bindFramebuffer(GL.FRAMEBUFFER, fb);
        if (fullViewPort) {            
            GL.viewport(0,0,W,H);
            //GL.clear(GL.COLOR_BUFFER_BIT | GL.DEPTH_BUFFER_BIT);
        } else {            
            GL.viewport(0,0,X_SIZE, Y_SIZE);
            var cc=shader_program.clearColor;
            GL.clearColor(cc[0], cc[1], cc[2], cc[3]);
            //GL.clearColor(1,0,0,1);
            GL.clear(GL.COLOR_BUFFER_BIT);
            GL.viewport(X,Y,W,H);
        }   
        
        if (setUniforms) setUniforms(shader_program);
        GL.bindTexture(GL.TEXTURE_2D, texture);
        GL.bindBuffer(GL.ARRAY_BUFFER, VBO) ;
        GL.vertexAttribPointer(shader_program.position, 2, GL.FLOAT, false,8,0) ;
        GL.bindBuffer(GL.ELEMENT_ARRAY_BUFFER, VBOI);
        GL.drawElements(GL.TRIANGLES, 6, GL.UNSIGNED_SHORT, 0);

        GL.flush();
    }

    var makeTexture=function(image){
        var texture=GL.createTexture();
        GL.pixelStorei(GL.UNPACK_FLIP_Y_WEBGL, true);
        GL.bindTexture(GL.TEXTURE_2D, texture);        
        GL.texImage2D(GL.TEXTURE_2D, 0, GL.RGBA, GL.RGBA, GL.UNSIGNED_BYTE, image);
        GL.texParameteri( GL.TEXTURE_2D, GL.TEXTURE_WRAP_S, GL.CLAMP_TO_EDGE );
        GL.texParameteri( GL.TEXTURE_2D, GL.TEXTURE_WRAP_T, GL.CLAMP_TO_EDGE );
        GL.texParameteri(GL.TEXTURE_2D, GL.TEXTURE_MAG_FILTER, GL.LINEAR);
        GL.texParameteri(GL.TEXTURE_2D, GL.TEXTURE_MIN_FILTER, GL.LINEAR);
        GL.bindTexture(GL.TEXTURE_2D, null);
        return texture;
    }

    return {
        init: function(spec) {
            SPEC=spec;
            if (!spec.gl) { JQCV=$("<canvas></canvas>");
            CV=JQCV.get(0); }
            GL = (spec.gl)?spec.gl:CV.getContext("experimental-webgl", {antialias: false, preserveDrawingBuffer: true});
            //create shaders
            
            get_shaders("shaders/bw_vertex.gl", "shaders/bw_fragment.gl", "bw", function(shader_program){
                shader_program.sampler = GL.getUniformLocation(shader_program, "sampler");
                shader_program.position = GL.getAttribLocation(shader_program, "position");
                GL.enableVertexAttribArray(shader_program.position);
                GL.useProgram(shader_program);
                GL.uniform1i(shader_program.sampler, 0);                
            }, [1,1,1,1]);
            
            get_shaders("shaders/ao_vertex.gl", "shaders/ao_fragment.gl", "ao", function(shader_program){
                shader_program.sampler = GL.getUniformLocation(shader_program, "sampler");
                shader_program.wh=GL.getUniformLocation(shader_program, "wh");
                shader_program.position = GL.getAttribLocation(shader_program, "position");
                GL.enableVertexAttribArray(shader_program.position);
                GL.useProgram(shader_program);
                GL.uniform1i(shader_program.sampler, 0);                
            }, [1,1,1,1]);

            get_shaders("shaders/normal_vertex.gl", "shaders/normal_fragment.gl", "normal", function(shader_program){
                shader_program.sampler = GL.getUniformLocation(shader_program, "sampler");
                shader_program.wh=GL.getUniformLocation(shader_program, "wh");
                
                shader_program.position = GL.getAttribLocation(shader_program, "position");
                GL.enableVertexAttribArray(shader_program.position);
                GL.useProgram(shader_program);
                GL.uniform1i(shader_program.sampler, 0);                
            }, [0.5,0.5,1,1]);
            
            //create VBOs
            VBO= GL.createBuffer();
            GL.bindBuffer(GL.ARRAY_BUFFER, VBO);
            GL.bufferData(GL.ARRAY_BUFFER,
	  		  new Float32Array([1,1,
                                            1,-1,
                                            -1,-1,
                                            -1,1]),
			  GL.STATIC_DRAW);

            VBOI=GL.createBuffer();
            GL.bindBuffer(GL.ELEMENT_ARRAY_BUFFER, VBOI);
            GL.bufferData(GL.ELEMENT_ARRAY_BUFFER,
	  		  new Uint16Array([0,2,1, 0,3,2]),
			  GL.STATIC_DRAW);
                        
           //render to texture BW
            FB=GL.createFramebuffer();
            GL.bindFramebuffer(GL.FRAMEBUFFER, FB);

            RB=GL.createRenderbuffer();
            GL.bindRenderbuffer(GL.RENDERBUFFER, RB);
           
            TEXTURE_BW=GL.createTexture();
            GL.bindTexture(GL.TEXTURE_2D, TEXTURE_BW);
            GL.texParameteri(GL.TEXTURE_2D, GL.TEXTURE_MAG_FILTER, GL.LINEAR);
            GL.texParameteri(GL.TEXTURE_2D, GL.TEXTURE_MIN_FILTER, GL.LINEAR);

            GL.texParameteri( GL.TEXTURE_2D, GL.TEXTURE_WRAP_S, GL.CLAMP_TO_EDGE );
            GL.texParameteri( GL.TEXTURE_2D, GL.TEXTURE_WRAP_T, GL.CLAMP_TO_EDGE );
           
            GL.framebufferTexture2D(GL.FRAMEBUFFER, GL.COLOR_ATTACHMENT0, GL.TEXTURE_2D, TEXTURE_BW, 0);
            GL.framebufferRenderbuffer(GL.FRAMEBUFFER, GL.DEPTH_ATTACHMENT, GL.RENDERBUFFER, RB);
            
            //render to texture AO
            FB_AO=GL.createFramebuffer();
            GL.bindFramebuffer(GL.FRAMEBUFFER, FB_AO);

            RB2=GL.createRenderbuffer();
            GL.bindRenderbuffer(GL.RENDERBUFFER, RB2);

            TEXTURE_AO=GL.createTexture();
            GL.bindTexture(GL.TEXTURE_2D, TEXTURE_AO);
            GL.texParameteri(GL.TEXTURE_2D, GL.TEXTURE_MAG_FILTER, GL.LINEAR);
            GL.texParameteri(GL.TEXTURE_2D, GL.TEXTURE_MIN_FILTER, GL.LINEAR);

            GL.texParameteri( GL.TEXTURE_2D, GL.TEXTURE_WRAP_S, GL.CLAMP_TO_EDGE );
            GL.texParameteri( GL.TEXTURE_2D, GL.TEXTURE_WRAP_T, GL.CLAMP_TO_EDGE );            

            GL.framebufferTexture2D(GL.FRAMEBUFFER, GL.COLOR_ATTACHMENT0, GL.TEXTURE_2D, TEXTURE_AO, 0);
            GL.framebufferRenderbuffer(GL.FRAMEBUFFER, GL.DEPTH_ATTACHMENT, GL.RENDERBUFFER, RB2);
            
            //render to texture normal
            FB_NORMAL=GL.createFramebuffer();
            GL.bindFramebuffer(GL.FRAMEBUFFER, FB_NORMAL);            

            TEXTURE_NORMAL=GL.createTexture();
            GL.bindTexture(GL.TEXTURE_2D, TEXTURE_NORMAL);
            GL.texParameteri(GL.TEXTURE_2D, GL.TEXTURE_MAG_FILTER, GL.LINEAR);
            GL.texParameteri(GL.TEXTURE_2D, GL.TEXTURE_MIN_FILTER, GL.LINEAR);

            GL.texParameteri( GL.TEXTURE_2D, GL.TEXTURE_WRAP_S, GL.CLAMP_TO_EDGE );
            GL.texParameteri( GL.TEXTURE_2D, GL.TEXTURE_WRAP_T, GL.CLAMP_TO_EDGE );            

            GL.framebufferTexture2D(GL.FRAMEBUFFER, GL.COLOR_ATTACHMENT0, GL.TEXTURE_2D, TEXTURE_NORMAL, 0);
            GL.framebufferRenderbuffer(GL.FRAMEBUFFER, GL.DEPTH_ATTACHMENT, GL.RENDERBUFFER, RB2);

            GL.bindTexture(GL.TEXTURE_2D, null);
            GL.bindRenderbuffer(GL.RENDERBUFFER, null);
            GL.bindFramebuffer(GL.FRAMEBUFFER, null);            

        },

        build: function(image,x,y,w,h){
            X=x, Y=y, W=image.width, H=image.height;
            //var w=image.width, h=image.height, result={};
            if (!SPEC.gl){
                JQCV.width(w);
                JQCV.height(h);
            }
            if (w==0){
                X_SIZE=W;
                w=W;
            }
            if (h==0){
                Y_SIZE=H;
                h=H;
            }
            if (!SPEC.gl) CV.width=w, CV.height=h;
            
            //render to texture
            GL.bindRenderbuffer(GL.RENDERBUFFER, RB);
            GL.renderbufferStorage(GL.RENDERBUFFER,GL.DEPTH_COMPONENT16 , W, H); //GL.RGBA4
            GL.bindTexture(GL.TEXTURE_2D, TEXTURE_BW);
            GL.texImage2D(GL.TEXTURE_2D, 0, GL.RGBA, W, H, 0, GL.RGBA, GL.UNSIGNED_BYTE, null);

            GL.bindRenderbuffer(GL.RENDERBUFFER, RB2);
            GL.renderbufferStorage(GL.RENDERBUFFER,GL.DEPTH_COMPONENT16 , X_SIZE, Y_SIZE); //GL.RGBA4
            GL.bindTexture(GL.TEXTURE_2D, TEXTURE_AO);
            GL.texImage2D(GL.TEXTURE_2D, 0, GL.RGBA, X_SIZE, Y_SIZE, 0, GL.RGBA, GL.UNSIGNED_BYTE, null);
            
            GL.bindTexture(GL.TEXTURE_2D, TEXTURE_NORMAL);
            GL.texImage2D(GL.TEXTURE_2D, 0, GL.RGBA, X_SIZE, Y_SIZE, 0, GL.RGBA, GL.UNSIGNED_BYTE, null);

            TEXTURE=makeTexture(image);
            
            Engrave.update(false);

            return {
                texture_ao: TEXTURE_AO,
                texture_normal: TEXTURE_NORMAL,
                texture_bw: TEXTURE_BW,
                texture_orig: TEXTURE
            }
        },

        update: function(image) {
            if (image){
                GL.bindTexture(GL.TEXTURE_2D, TEXTURE);
                GL.texImage2D(GL.TEXTURE_2D, 0, GL.RGBA, GL.RGBA, GL.UNSIGNED_BYTE, image);
            }

            GL.disable(GL.DEPTH_TEST);
            makeImage(SHADER_PROGRAMS.bw, TEXTURE, null, true, FB);

            makeImage(SHADER_PROGRAMS.ao, TEXTURE_BW,
                function(shader_program){
                    GL.uniform2f(shader_program.wh, W,H);
                }, false, FB_AO);;

            makeImage(SHADER_PROGRAMS.normal, TEXTURE_BW,
                function(shader_program){
                    GL.uniform2f(shader_program.wh, W,H);
                }, false, FB_NORMAL);

            GL.bindFramebuffer(GL.FRAMEBUFFER, null);
            GL.viewport(0,0,window.innerWidth, window.innerHeight);
            GL.enable(GL.DEPTH_TEST);

        },


        test: function(){

        }

    }

})();

