var SETTINGS={
    view3D: {
        easing: 0.9,
        sensitivity: 0.001,
        realTime: true
    }
};


var Model=(function() {
    return {
        load: function(jsonFile){
            $.ajax({
                url: jsonFile,
                dataType: "json",
                context: document.body,
                async: true
            }).success(function(data){
                Scene.loadModelCollada(data.model3D, data.materials);

            });
        }
    }
})();


/*
 * spec.renderer : three renderer
 */
var Scene=(function(){
    var threeScene=false, camera=false, renderer=false;
    var ambientLight, directionalLight;
    var bgScene, bgCam;

    var animate=function() {
        requestAnimationFrame(animate);
        Navigation.animate();
        renderer.clear();
        renderer.render(bgScene, bgCam);
        renderer.render(threeScene, camera);
    }

    return {
        init: function(spec){
            renderer=spec.renderer;
            renderer.autoClear = false;

            //make the scene
            threeScene=new THREE.Scene();

            //set the camera
            camera = new THREE.PerspectiveCamera(35, window.innerWidth / window.innerHeight, 1, 50);
            camera.position.set(0, 0, 20);
            threeScene.add(camera);
            THREEx.WindowResize.bind(renderer, camera);

            //set lighting
            ambientLight = new THREE.AmbientLight( 0xffffff );
            threeScene.add( ambientLight );
            directionalLight = new THREE.DirectionalLight( 0xffffff );
            directionalLight.position.set(0,0,1);
            threeScene.add( directionalLight );

            //set background
            var backgroundTexture=new THREE.ImageUtils.loadTexture("images/textures/background.png");
            var bg = new THREE.Mesh(
                new THREE.PlaneGeometry(2, 2, 0),
                new THREE.MeshBasicMaterial({map: backgroundTexture})
            );
            bg.material.depthTest = false,
            bg.material.depthWrite = false;
            bgScene = new THREE.Scene(),
            bgCam = new THREE.Camera();
            bgScene.add(bgCam);
            bgScene.add(bg);

        },

        loadModelJS: function(modelURL){
            var loader = new THREE.JSONLoader();
            loader.load(modelURL, function (geom) {
                var mat=new THREE.MeshLambertMaterial( { color: 0xFF0000 } ); //new THREE.MeshFaceMaterial();
                var model=new THREE.Mesh(geom, mat);
                threeScene.add(model);                
                animate();
            });
        },

        loadModelCollada: function(modelURL, materials){
           
            var loader = new THREE.ColladaLoader();
            loader.load(modelURL, function (dae) {
                var model=dae.scene;
                //override materials if necessary                
                model.children.map(function(mesh){
                    var l=_.where(materials, {mesh: mesh.name});                    
                    if (l.length==0) return;
                    var mat=l[0];
                    if (mat.cubeMap && !mat.engraveImage){
                        //mesh.material=new THREE.MeshBasicMaterial( { color: parseInt(mat.color), envMap: MatLib.get_textureCube() } )
                        mesh.material = new THREE.MeshPhongMaterial( {
					color: parseInt(mat.diffuse || 0xffffff),
					specular: parseInt(mat.specular || 0xffffff),
					ambient: parseInt(mat.ambient || 0xffffff),
					shininess: parseFloat(shininess || 1),
					envMap: MatLib.get_textureCube(),
					reflectivity: parseFloat(mat.reflectivity || 0.1)
				} );
                                
                    }
                    if (mat.doubleSided){
                        mesh.material.side=THREE.DoubleSide;
                    }
                    if (mat.engraveImage){
                        var ambient = parseInt(mat.ambient), diffuse = parseInt(mat.diffuse), specular = parseInt(mat.specular), shininess = parseFloat(mat.shininess);

                        // normal map shader
                        var shader = THREE.ShaderLib[ "normalmap" ];
                        var uniforms = THREE.UniformsUtils.clone( shader.uniforms );

                        uniforms[ "enableAO" ].value = true;
                        uniforms[ "enableDiffuse" ].value = false;
                        uniforms[ "enableSpecular" ].value = false;
                        uniforms[ "enableReflection" ].value = true;
                        uniforms[ "enableDisplacement" ].value = false;

                        var texture_normal=MatLib.get_emptyTexture(),
                            texture_ao=MatLib.get_emptyTexture();
                            //texture_displacement=MatLib.get_emptyTexture3();

                        mat.texture_ao=texture_ao,
                        mat.texture_normal=texture_normal;
                        
                        //texture_ao=THREE.ImageUtils.loadTexture("images/textures/testAO.jpg");
                        
                        uniforms[ "tNormal" ].value = texture_normal;
                        uniforms[ "tAO" ].value =texture_ao;
                        //uniforms[ "tDisplacement" ].value = texture_displacement;

                        uniforms[ "uDisplacementBias" ].value = - 0.428408;
                        uniforms[ "uDisplacementScale" ].value = 2.436143;

                        uniforms[ "uNormalScale" ].value.y = -1;

                        uniforms[ "uDiffuseColor" ].value.setHex( diffuse );
                        uniforms[ "uSpecularColor" ].value.setHex( specular );
                        uniforms[ "uAmbientColor" ].value.setHex( ambient );

                        uniforms[ "uShininess" ].value = shininess;

                        uniforms[ "tCube" ].value = MatLib.get_textureCube();
                        uniforms[ "uReflectivity" ].value = mat.reflectivity;

                        uniforms[ "uDiffuseColor" ].value.convertGammaToLinear();
                        uniforms[ "uSpecularColor" ].value.convertGammaToLinear();
                        uniforms[ "uAmbientColor" ].value.convertGammaToLinear();

                        var parameters = { fragmentShader: shader.fragmentShader, vertexShader: shader.vertexShader, uniforms: uniforms, lights: true, fog: false };
                        mesh.geometry.computeTangents();
                        mesh.material = new THREE.ShaderMaterial( parameters );
                        
                        
                        var engraveImage=new Image();
                        engraveImage.src=mat.engraveImage;
                        engraveImage.onload=function(e){ 
                            setTimeout(function() {                                
                            
                            mat.originalWidth=engraveImage.width;
                            mat.originalHeight=engraveImage.height;
                            
                            var engraved=Engrave.build(engraveImage, 0,0, 0,0, false);
                            var t_normal=mat.texture_normal;
                            t_normal.generateMipmaps=false;
                            t_normal.__webglTexture=engraved.texture_normal;

                            var t_ao=mat.texture_ao;
                            t_ao.generateMipmaps=false;
                            t_ao.__webglTexture=engraved.texture_ao;
                            }, 10);
                        }
                    }
                    if (mat.customizable) {
                        MatForms.make(mat, {
                            normal: texture_normal,
                            ao: texture_ao
                        })
                    }
                });

                threeScene.add(model);
                Navigation.set_model(model);
                window.model=model; window.camera=camera;
                animate();
            });
        }

    }
})();


var View3D=(function(){
    var renderer, canvas, container, jqCanvas;
    return {
        init: function(){
            container=$("#View3D");
            renderer = new THREE.WebGLRenderer({
                antialias : true,
            });
            renderer.setClearColor(0xBBBBBB,1);
            renderer.gammaInput = true,
            renderer.gammaOutput = true;
            renderer.setSize( window.innerWidth, window.innerHeight );
            canvas=renderer.domElement;
            jqCanvas=$(canvas);
            container.append(canvas);
            
            MatLib.init();
            Scene.init({renderer: renderer});
            Navigation.set(jqCanvas);
            
        },
        get_gl: function(){
            return renderer.context;
        }
    }
})();

var Navigation=(function() {
    var model=false, container=false;
    var mouseX=0, mouseY=0, drag=false, dX=0, dY=0, theta=0, phi=0;

    var rotateModel=function() {
        if (!model) return;
        theta+=dX*SETTINGS.view3D.sensitivity, phi+=0.7*dY*SETTINGS.view3D.sensitivity*window.innerWidth/window.innerHeight;
        model.rotation.y=theta,
        model.rotation.x=phi;
        model.updateMatrix();
    }

    var mouseDown=function(e) {
            mouseX=e.pageX,
            mouseY=e.pageY;
            drag=true;
            e.preventDefault();
            e.stopPropagation();
        },
    mouseUp=function(e) {
            drag=false;
            //e.preventDefault();
            //e.stopPropagation();
        },
    mouseOut=function(e){
            drag=false;
        },
    mouseMove=function(e){
            if (!drag || !model) return false;
            dX=e.pageX-mouseX;
            dY=e.pageY-mouseY;
            mouseX=e.pageX; mouseY=e.pageY;
            rotateModel();
            e.preventDefault();
            e.stopPropagation();
        },

     mouseWheel=function(e, delta) {
         e.preventDefault();
         e.stopPropagation();
         return false;
     }

    return {
        set: function(cv) {
            container=cv;
            container.mousedown(mouseDown).mouseup(mouseUp).mouseout(mouseOut).mousemove(mouseMove).mousewheel(mouseWheel);
        },

        unset: function() {
            container.unbind("mousedown", mouseDown).unbind("mouseup", mouseUp).unbind("mouseout", mouseOut).unbind("mousemove", mouseMove).unbind("mousewheel", mouseWheel);
        },

        set_model: function(m){
            model=m;
        },

        animate: function() {
            if (drag) return;
            dX*=SETTINGS.view3D.easing,
            dY*=SETTINGS.view3D.easing;
            rotateModel();
        }
    }
})();


var MatLib=(function(){
    var textureCube=false, emptyTexture, emptyTexture2, emptyTexture3;
    return {
        init: function(){
            //cube texture
            var r = "images/textures/cube/",
                ext=".png";
            var urls = [ r + "posx"+ext, r + "negx"+ext,
                         r + "posy"+ext, r + "negy"+ext,
                         r + "posz"+ext, r + "negz"+ext ];
            textureCube = THREE.ImageUtils.loadTextureCube( urls );            
                        
            
        },

        get_textureCube: function() {
            return textureCube;
        },
        
        get_emptyTexture: function() {
            return new THREE.Texture();
        }
    }
})();

var MatForms=(function() {
    var getImageForm=function(matForm, mat, textures){
        var fileInput=$("<button></button").appendTo(matForm);
        
        fileInput.text(LANG[PHPPARAMS.lang].browse).button().width(200).file().choose(function(e, input) {
                                        var reader = new FileReader();
                                        
                                        reader.onload = function(e) {
                                            var image=$('<img/>').load(function() {
                                                var i2D=Image2D.instance({image: image.get(0), mat: mat});
                                                
                                                //build preview
                                                if (mat.preview) mat.preview.remove();
                                                mat.preview=i2D.buildPreview(matForm, 200);

                                                //build texture image
                                                var textureCv=i2D.getReduced(mat.width, mat.height, true, false);

                                                var texture=new Image;
                                                texture.src=textureCv.toDataURL("image/png");
                                                texture.onload=function(){                                                    
                                                    var engraved=Engrave.build(texture, mat.left, mat.top, mat.originalWidth, mat.originalHeight, false);                                                    
                                                }
                                            }).attr('src', e.target.result);     
                                        };
                                        reader.readAsDataURL(input[0].files[0]);                                        
				});            
    }
    
    return {
        make: function(mat, textures) {
            var matForm=$("<div></div>").attr("class", "matForm").appendTo("#matFormsContainer");
            
            switch(mat.customType){
                case "image":
                    matForm.append(getImageForm(matForm, mat, textures));
                    break;
            }
            
            
        }
        
    }
    
    
})();
/*
 * spec.image : javascript image object
 * spec.mat: mat
 */
var Image2D=(function(){
    return {
        instance: function(spec) {
            var image=spec.image,
                width=spec.image.width,
                height=spec.image.height,
                ratio=spec.image.width/spec.image.height;
            
            var fullCv=document.createElement("canvas");
            fullCv.width=spec.mat.width,
            fullCv.height=spec.mat.height;
            var fullCtx=fullCv.getContext("2d");
            fullCtx.fillStyle="rgb(255,255,255)";

            var that={
                getReduced: function(w, h, fill, crop){
                    
                    //console.log(w,h);
                    var cv=document.createElement("canvas");
                    var sx=w/width, sy=h/height;
                    var s=Math.min(sx, sy);
                    var x,y;
                    if (sx<sy){
                        x=0,
                        y=(h-s*height)/2;
                    } else {
                        y=0;
                        x=(w-s*width)/2;
                    }
                    if (crop){
                        //cv.trans_x=x,
                        //cv.trans_y=y;
                        w=s*width, h=s*height;x=0; y=0;
                    }

                    cv.width=w; cv.height=h;
                    var ctx=cv.getContext("2d");

                    if (fill) {
                        ctx.fillStyle="rgb(255,255,255)";
                        ctx.fillRect(0,0,w,h);
                    }
                    ctx.drawImage(image,x,y, s*width,s*height);
                    
                    return cv;
                },

                buildPreview: function(jqContainer, width){
                    var w=spec.mat.width,
                        h=spec.mat.height;
                    var zoom=1, trans_x=0, trans_y=0;

                    var a=h/w;
                    var height=Math.round(a*width); //height of the preview
                    var previewImg=that.getReduced(width, height);

                    var previewContainer=$("<div></div>").appendTo(jqContainer);
                    var jqcv=$("<canvas></canvas>");
                    var cv=jqcv.get(0);
                    cv.width=width, cv.height=height;
                    var ctx=cv.getContext("2d");                    
                    
                    previewContainer.append(jqcv);

                    ctx.fillStyle="rgb(255,255,255)";
                    var updateCV=function() {
                        ctx.save();
                        ctx.fillRect(0,0,width, height);                        
                        ctx.translate(width/2+trans_x,height/2+trans_y);
                        ctx.scale(zoom, zoom);
                        ctx.drawImage(previewImg,-previewImg.width/2, -previewImg.height/2);
                        ctx.restore();
                    }

                    var buildRealScale=function(){
                        var fullScale=fullCv.width/width;
                        fullCtx.fillRect(0,0,fullCv.width, fullCv.height);
                        fullCtx.save();
                        fullCtx.translate(fullCv.width/2+fullScale*trans_x,fullCv.height/2+fullScale*trans_y); //move to center
                        fullCtx.scale(zoom, zoom);
                        fullCtx.drawImage(image,-fullScale*previewImg.width/2, -fullScale*width/(2*ratio), fullScale*width, fullScale*width/ratio);
                        fullCtx.restore();
                        //Debug.image(fullCv, true);
                        Engrave.update(fullCv);
                    }

                    var cursorOnImage=function(e){
                        var offset=jqcv.offset();
                        var x_min=width/2+trans_x-previewImg.width/2*zoom,
                            x_max=width/2+trans_x+previewImg.width/2*zoom,
                            y_min=height/2+trans_y-previewImg.height/2*zoom,
                            y_max=height/2+trans_y+previewImg.height/2*zoom;

                        var x=e.pageX-offset.left,
                            y=e.pageY-offset.top;
                        return (x>x_min && x<x_max && y>y_min && y<y_max);
                     }

                    var drag=false, oldX=0, oldY=0, cursorMove=false;
                    jqcv.mousedown(function(e) {
                        if (cursorOnImage(e)){
                            oldX=e.pageX,
                            oldY=e.pageY,
                            drag=true;
                        }
                    }).mouseup(function(e){
                        if (drag) buildRealScale();
                        drag=false;
                    }).mouseout(function(e){
                        drag=false;
                    }).mousemove(function(e){
                        if (drag){
                            var dX=e.pageX-oldX,
                                dY=e.pageY-oldY;

                            oldX=e.pageX,
                            oldY=e.pageY;
                            trans_x+=dX,
                            trans_y+=dY;
                            updateCV();
                            e.preventDefault();
                            if (SETTINGS.view3D.realTime) buildRealScale();
                            return false;
                        } else if (cursorOnImage(e)){
                            if (!cursorMove){
                                jqcv.css("cursor", "move")
                                cursorMove=true;
                            }
                        } else if (cursorMove){
                            jqcv.css("cursor", "default")
                            cursorMove=false;
                        }
                        e.stopPropagation();
                    })

                    updateCV();

                    var sliderZoom=$("<div></div>").width(width).appendTo(previewContainer);
                    sliderZoom.slider({
                        min: 0.5,
                        max: 2,
                        value: 1,
                        step: 0.001,

                        slide: function(event, ui){
                            zoom=ui.value;
                            updateCV();
                            if (SETTINGS.view3D.realTime) buildRealScale();
                        },

                        stop: function(event, ui){
                            buildRealScale();
                        }
                    })


                    return previewContainer;
                }
                
                
            }
            return that;
        }
        
        
    }
    
})();


var Debug=(function() {
    var initialized=false, debugDiv, oldImage=false;
    var init=function() {
        debugDiv=document.createElement("div");
        document.body.appendChild(debugDiv);
        debugDiv.style.position="absolute";
        debugDiv.style.top="0px";
        debugDiv.style.left="0px";
        initialized=true;
    }
    
    
    return {
        image: function(image, replace) {
            if (!initialized) init();
            if (oldImage && replace) {
                debugDiv.removeChild(oldImage);
                oldImage=image;
            }
            debugDiv.appendChild(image);
            
        }
    }
    
})();


var LANG={
    fr: {
        browse: "Change picture"
    },

    en: {
        browse: "Change picture"
    }

}



$(document).ready(function(){    
    View3D.init();
    Engrave.init({root: "js/engrave/", gl: View3D.get_gl()});
    Model.load("models/zippo.json");
    
});


var PHPPARAMS={lang: "fr"};